/*
 * Example of sorting an array in-place
 * using Fork-Join parallelism.
 *
 * This demo illustrates:
 * - fork-join parallelism / divide-and-conquer with C++ threads
 * 
 * Fork and join both allow threads to safely communicate
 * with each other and implicitly pass ownership of data.
 *
 * CSE 374, 2015-03 BM
 */

#include <iostream>
#include <thread>
#include <cstdlib>
#include <cstring>

// note that this function is not atomic; so print outs may be interleaved from different threads
void print_array(int* arr, int len) {
  std::cout << "[";
  for (int i=0; i<len; i++) {
    std::cout << arr[i] << ",";
  }
  std::cout << "]\n";
}

// sort array in-place using recursive mergesort and fork-join parallelism
void sort(int* array, int len) {
  
  // base case: if array is length 1, then it is already sorted
  if (len == 1) {
    return;
  }
  
  // split the array into two roughly equal halves
  int* left_arr = array;
  int left_len = len/2;
  int* right_arr = array + len/2;
  int right_len = len - len/2;

  // fork: the current thread creates two new threads,
  // one to sort each half of array.
  std::thread left(sort, left_arr, left_len);
  std::thread right(sort, right_arr, right_len);

  /* If the current thread were to read or write the array
   * right here (i.e., before the following joins complete)
   * then it would be a "data race" (type of concurrency bug
   */

  // join: the current thread waits for both child threads
  // to finish
  left.join();
  right.join();

  // Now that the child threads are done, the current thread
  // can safely touch the array again.

#ifdef DEBUG
  std::cout << "merging: " << array << std::endl;
  print_array(left_arr, left_len);
  print_array(right_arr, right_len);
#endif

  /* merge the two results */

  // create a temporary array to put output of merge
  int* temp = new int[len];
  int l = 0;
  int r = 0;
  int m = 0;
  while (l < left_len && r < right_len) {
    if (left_arr[l] <= right_arr[r]) {
      temp[m] = left_arr[l];
      l++;
    } else {
      temp[m] = right_arr[r];
      r++;
    }
    m++;
  }

  // copy the rest of the left (if right ran out of elements in the first while)
  while (l < left_len) {
    temp[m] = left_arr[l];
    l++;
    m++;
  }
  
  // OR copy the rest of the right (if left ran out of elements in the first while)
  while (r < right_len) {
    temp[m] = right_arr[r];
    r++;
    m++;
  }

  // copy the merged data in temp into array
  memcpy(array, temp, len*sizeof(int));
  delete[] temp;
  
#ifdef DEBUG
  std::cout << "done: " << array << std::endl;
  print_array(array, len);
#endif

  // array is now sorted
}


int main() {
  const int len = 20;

  // initialize the array to be sorted
  int* arr = new int[len];
  for (int i=0; i<len; i++) {
    arr[i] = rand() % (len*2);
  }

  print_array(arr, len);
  sort(arr, len);
  print_array(arr, len);

  delete[] arr;
}
